/* 
 * readroff.c --
 *     This file implements simple extension to convert nroff output into
 *     Tk text widget tags
 *
 *
 *  Copyright (c) 1997 by Victor B. Wagner
 * 
 * This file is distributed under Gnu Public License.
 * Contact Free Software Foundation to obtain full text of license
 */

#include <tcl.h>

 static char rcs_id[]="$Header$";
/*
 * For compilation with BC 4.x
 */
#ifdef __unix
# define _export
#endif

/* constants */

/*
 * Types of Tags for interrnal use
 */
#define NORMAL 0
#define BOLD 1
#define UNDERLINE 2

/*
 * Default buffersize and buffer increment, if size not sufficient
 */

#define STARTSIZE 1024
#define SIZEDELTA 256

/*
 * Global variables
 */

int run_length, /* amount of characters in current run 
                  (sequence of symbols with current attribute)
                 */
    mode; /* tag code, applicable to current character */
char *result, /* Dynamic array to hold formatted output */
     *limit; /* Pointer to last allocated element of result[] */


/*
 * Inserts raw symbol into result buffer at position *dest, reallocating
 * buffer, if neccessary
 */

void inschar(char **dest,int c)
{
    if (*dest==limit) {
       int size,pos;
       size=limit-result+SIZEDELTA;
       pos=*dest-result;
       result=Tcl_Realloc(result,size+1);
       limit=result+size;
       *dest=result+pos;
    }
    *((*dest)++)=c;
}


/*
 * Adds character from input to output buffer,
 * taking care of quoting special (for Tcl) symbols
 */

void addchar(char **dest,int c)
{
    static char escaped[]={"{}\"\\[]$"}; 
    if (strchr(escaped,c)) {
        inschar(dest,'\\');
    };
    inschar(dest,c);
    run_length++; 
}

/*
 * Closes the run, converting into valid Tcl list element
 * and adds tag name after it
 */

void close_mode(char **dest)
{
   char *modenames[]={"roff_text","roff_bold","roff_udl"};
   char *src;
   inschar (dest,'\"');
   inschar (dest,' ');
   for (src=modenames[mode];*src;inschar(dest,*(src++)));
}

/* 
 * New character sequence with mode newmode
 * Closes previous, if neccessary
 */
void setmode(char **dest,int newmode)
{
   if (run_length) {
      close_mode(dest);
      /* put space between tag and next run and opening quote for this
              next run */
      inschar(dest,' '); 
      inschar(dest,'\"');
   }
   mode=newmode;
}

/*
 * Readroff_Command
 *   implements Tcl Command Readroff
 *  input:
 *      - argv[1] should contain string, outputted from nroff(1)
 *  Output:
 *      Tcl list with interchanging pieces of text and tags describing text
 *         appearance 
 *  Side effects: None
 */
int _export Readroff_Command(ClientData data, Tcl_Interp *interp,int argc, 
                       char **argv)
{ 
   int prevchar;
   char *src,*dest;
   if (argc!=2) { 
      Tcl_SetResult(interp,"Wrong # of args. Should be readroff string",
               TCL_STATIC);
      return TCL_ERROR;
   }
   result=Tcl_Alloc(STARTSIZE+1);
   limit=result+STARTSIZE;
   prevchar=0;
   dest=result;
   run_length=0;
   *(dest++)='\"';
   mode=NORMAL; 
   for(src=argv[1];*src;src++) {
      if (*(src+1)==010) { 
         prevchar=*src; 
      } else {
          if (*src!=010) {
              if (prevchar) {
                  if (prevchar=='_'||*src=='_') { 
                     switch (mode) {
                       case BOLD: if(prevchar=='_'&&*src=='_') 
                                  addchar(&dest,'_'); 
                                  break;
                       case NORMAL: setmode(&dest,UNDERLINE);
                       case UNDERLINE: if (prevchar!='_') {
                                          addchar(&dest,prevchar);
                                       } else {
                                          addchar(&dest,*src);
                                       }
                     }
                  }   
                  prevchar=0;
              } else {
                 if (prevchar==*src) {   
                     if (mode != BOLD ) {
                         setmode (&dest,BOLD);
                     }
                     addchar(&dest,*src);
                     prevchar=0;
                 } else { 
                    Tcl_Free(result); 
                    Tcl_SetResult(interp,"Invalid nroff formatting found",
                           TCL_STATIC);
                    return TCL_ERROR;
                 }
             }
         } else { 
            if (mode!=NORMAL) { 
                setmode(&dest,NORMAL);
            }
            addchar(&dest,*src);    
         }
     }
  }/*loop*/
  close_mode(&dest);
  *dest=0;
  Tcl_SetResult(interp,result,TCL_DYNAMIC);
  return TCL_OK;
}    


/* 
 * Empty procedure, used as TclDeleteCmdProc
 */
void do_nothing(ClientData data) {} 

/* 
 * Readroff_Init -
 *   Initilazes package readroff.
 * Return : TCL_OK if successiful
 * Side effects : registers Tcl command readroff
 *                provides package readroff (if Tcl > 7.5
 */
                 
int _export Readroff_Init(Tcl_Interp *interp)
{
  if (!Tcl_CreateCommand(interp,"readroff",Readroff_Command,NULL,do_nothing)) {
      return  TCL_ERROR;
  }
  return Tcl_PkgProvide(interp,"readroff","0.1");
}
